#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <math.h>
#include "mtxutl.h"

#define PI 3.14159265358979323846
#define END_OF_VEC -1

#define NKOUHO   20
#define NKOUHO_LONG   500

#define MAX(X,Y)    ( ((X)>(Y))?(X):(Y) )
#define MIN(X,Y)    ( ((X)<(Y))?(X):(Y) )
